############################################################################
##
## Copyright (C) 2006-2007 University of Utah. All rights reserved.
##
## This file is part of VisTrails.
##
## This file may be used under the terms of the GNU General Public
## License version 2.0 as published by the Free Software Foundation
## and appearing in the file LICENSE.GPL included in the packaging of
## this file.  Please review the following to ensure GNU General Public
## Licensing requirements will be met:
## http://www.opensource.org/licenses/gpl-license.php
##
## If you are unsure which license is appropriate for your use (for
## instance, you are interested in developing a commercial derivative
## of VisTrails), please contact us at vistrails@sci.utah.edu.
##
## This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
## WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
##
############################################################################
import os
import os.path
import sys
import platform
import popen2
from core.utils import unimplemented, VistrailsInternalError
import core.requirements

################################################################################

systemType = platform.system()

if systemType in ['Windows', 'Microsoft']:
    from core.system.windows import guessTotalMemory, temporaryDirectory, \
        homeDirectory, remoteCopyProgram, remoteShellProgram, \
        graphVizDotCommandLine, removeGraphvizTemporaries, link_or_copy, \
        executable_is_in_path, TestWindows

elif systemType in ['Linux']:
    from core.system.linux import guessTotalMemory, temporaryDirectory, \
        homeDirectory, remoteCopyProgram, remoteShellProgram, \
        graphVizDotCommandLine, removeGraphvizTemporaries, link_or_copy, \
        XDestroyWindow, executable_is_in_path, TestLinux

elif systemType in ['Darwin']:
    from core.system.osx import guessTotalMemory, temporaryDirectory, \
        homeDirectory, remoteCopyProgram, remoteShellProgram, \
        graphVizDotCommandLine, removeGraphvizTemporaries, link_or_copy, \
        executable_is_in_path, TestMacOSX
else:
    print "Critical error"
    print "VisTrails could not detect your operating system."
    sys.exit(1)

# Makes sure root directory is sensible.
if __name__ == '__main__':
    _thisDir = sys.argv[0]
else:
    _thisDir = sys.modules[__name__].__file__
_thisDir = os.path.split(_thisDir)[0]
__rootDir = _thisDir + '/../../'

__dataDir = __rootDir + 'data/'

def setVistrailsDataDirectory(d):
    """ setVistrailsDataDirectory(d:str) -> None 
    Sets vistrails data directory taking into account environment variables

    """
    global __dataDir
    new_d = os.path.expanduser(d)
    new_d = os.path.expandvars(new_d)
    while new_d != d:
        d = new_d
        new_d = os.path.expandvars(d)
    __dataDir = d + '/'

def setVistrailsDirectory(d):
    """ setVistrailsDirectory(d:str) -> None 
    Sets vistrails root directory taking into account environment variables

    """

    global __rootDir
    new_d = os.path.expanduser(d)
    new_d = os.path.expandvars(new_d)
    while new_d != d:
        d = new_d
        new_d = os.path.expandvars(d)
    __rootDir = d + '/'

def visTrailsRootDirectory():
    """ visTrailsRootDirectory() -> str
    Returns vistrails root directory

    """
    return __rootDir

def vistrailsDirectory():
    """ vistrailsDirectory() -> str 
    Returns vistrails examples directory

    """
    return visTrailsRootDirectory() + '../examples/'

def packagesDirectory():
    """ packagesDirectory() -> str 
    Returns vistrails packages directory

    """
    return visTrailsRootDirectory() + 'packages/'

def blankVistrailFile():
    unimplemented()

def resourceDirectory():
    """ resourceDirectory() -> str 
    Returns vistrails gui resource directory

    """
    return visTrailsRootDirectory() + 'gui/resources/'

def defaultOptionsFile():
    """ defaultOptionsFile() -> str 
    Returns vistrails default options file

    """
    return homeDirectory() + "/.vistrailsrc"

def defaultDotVistrails():
    """ defaultDotVistrails() -> str 
    Returns VisTrails per-user directory.

    """
    return homeDirectory() + "/.vistrails"

def defaultBookmarksFile():
    """ defaultBookmarksFile() -> str
    Returns default Vistrails per-user bookmarks file

    """
    return defaultDotVistrails() + "/bookmarks.xml"

def pythonVersion():
   """pythonVersion() -> (major, minor, micro, release, serial)
Returns python version info."""
   return sys.version_info

def vistrailsVersion():
   """vistrailsVersion() -> string - Returns the current VisTrails version."""
   # 0.1 was the Vis2005 version
   # 0.2 was the SIGMOD demo version
   # 0.3 was the plugin/vtk version
   # 0.4 is cleaned up version with new GUI
   return '0.4'

def vistrailsRevision():
    """vistrailsRevision() -> str 
    When run on a working copy, shows the current svn revision else
    shows the latest release revision

    """
    release = "444"
    if core.requirements.executable_file_exists('svn'):
        if systemType not in ['Windows', 'Microsoft']:
            process = popen2.Popen4("svn info")
            result = -1
            while result == -1:
                result = process.poll()
            svn_output = process.fromchild
        else:
            #Popen4 does not seem to be present on Windows
            svn_output, input = popen2.popen4("svn info")
            result = 0
            
        revision_line = svn_output.readlines()[4][:-1].split(' ')
        if result == 0:
            if revision_line[0] == 'Revision:':
                return revision_line[1]
    return release
        
def aboutString():
   """aboutString() -> string - Returns the about string for VisTrails."""
   return """VisTrails version %s.%s -- vistrails@sci.utah.edu

Copyright (c) 2006-2007 University of Utah. All rights reserved.
http://www.vistrails.org

THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY \
APPLICABLE LAW. EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT \
HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM AS IS WITHOUT \
WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT \
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR \
A PARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY AND \
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE \
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR \
CORRECTION. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO \
IN WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY \
MODIFY AND/OR REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE \
TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR \
CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE \
PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING \
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A \
FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF \
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH \
DAMAGES.""" % (vistrailsVersion(), vistrailsRevision())

################################################################################

import unittest

if __name__ == '__main__':
    unittest.main()
